
; Include File for Encod and Decod Routines
; Version 1.02 - March, 2010
; by: Dave Sheldon - DedNDave

;############################################################################################

Encod   PROTO   :DWORD,:DWORD,:DWORD
Decod   PROTO   :DWORD,:DWORD,:DWORD

;############################################################################################

        .DATA

;############################################################################################
;
; Encode/Decode Hash Table
; Version 1.02 - March, 2010
; by: Dave Sheldon - DedNDave
;
; required by both Encod and Decod routines
;
; Notes:
;
; 1) The table should be defined in the initialized ".data" segment.
; 2) The same table is used for both encoding and decoding.
; 3) Strings encoded with one table cannot be decoded with a different table.
; 4) The table may contain any 64 ASCII characters desired, however they should all be
;    characters that are allowed in .INI files and/or registry entries as "data".
;    Characters that are interpreted as delimiters or special characters should not be
;    used. For example, "#@=:;[]",'\./" characters should be avoided. It may also be
;    prudent to avoid using "0", as some parsers may suppress leading zeros.
;    Likewise, the space character may be interpreted as a parameter delimiter.
; 5) Each of the 64 characters must appear in the table only once.
; 6) The characters may appear in any order, a fact that may be
;    used to enhance the random appearance of the encoded strings.
; 7) The Encod and Decod routines do not validate this table.
;    It is left to the programmer to insure it is correct.
;
; example character set used:
;           $-123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz
; this character set should be usable in .INI files and/or registry entries

_EncTbl db 'tNqABRs7$GC64KmzMOP3jZDg5vHIbfu_rXnohJWekFQVTESa1xY2wdyLl8-cU9pi'

;############################################################################################

        .CODE

;############################################################################################

        OPTION  PROLOGUE:NONE
        OPTION  EPILOGUE:NONE

Encod   PROC    szInputString:DWORD,lpOutputBuffer:DWORD,dwSize:DWORD
;
; Encode decimal ASCII string to hashed base 64 ASCII string with checksums
;
; accepts ASCII numeric string from 0 to 1023 bytes in length
; encode/compress/hash to base 64 ASCII string
; SUM checksum, XOR checksum, and length values are encoded and appended to the string
; to be used with the companion Decod routine
;
; Version 1.02 - March, 2010
; by: Dave Sheldon - DedNDave
;
;*********************************************************
;
; Use:
;
;       INVOKE  Encod,
;               szInputString,
;               lpOutputBuffer,
;               dwSize
;
;   szInputString = address of zero terminated ASCII numeric string (1023 bytes max. length)
;  lpOutputBuffer = address of output buffer
;          dwSize = size of output buffer in bytes
;
; Returns:
;
;  EAX = status code
;          0 = ERROR_SUCCESS
;              the string has successfully been encoded; ECX, EDX contain meaningful values
;         13 = ERROR_INVALID_DATA
;              indicates that a non-ASCII-numeric character was encountered
;         87 = ERROR_INVALID_PARAMETER
;              indicates a problem with the buffer address and/or size parameters
;        122 = ERROR_INSUFFICIENT_BUFFER
;              the output buffer is too small to accommodate the resultant encoded string
;        234 = ERROR_MORE_DATA
;              input string is longer than the routine was designed to handle (1023 bytes)
;
;  ECX = output string length, not including null terminator
;        if EAX = 0 (no error)
;           ECX = length of encoded string, not including null terminator
;        if EAX <> 0 (error)
;           ECX = 0
;
;  EDX = address of output buffer
;        if EAX = 0 (no error)
;           EDX = address of output buffer
;        if EAX <> 0 (error)
;           EDX = address of output buffer (null string)
;
; Notes:
;
; 1) The _EncTbl is required.
; 2) The original numeric string is not altered during operation.
; 3) The minimum output buffer size is 7 bytes, and must include room for a null terminator.
; 4) The resulting output string is left-justified in the output buffer field.
; 5) The required size of the output buffer may be calculated as follows:
;
;    ********************************************************************
;    IMPORTANT: This formula is preliminary. It needs to be tested.
;               For now, use it as an ESTIMATE of the ouptut buffer size.
;    ********************************************************************
;
;    Dmax = Floor(N x Log64(10)) + 7
;    Log64(10) = Log(10)/Log(64) = 0.55365468248122705797838657158157
;
;    where D = number of bytes in output buffer (including null terminator)
;          N = number of input digits (not including null terminator)
;
;    Example: For a numeric input string of 81 digits...
;
;    Dmax = Floor(81 x 0.55365468248122705797838657158157) + 7
;         = 51
;    a 51 byte output buffer is required, which includes the null string terminator
;
;    This calculation includes the encoded checksum and length
;    values, as well as a null string terminator byte.
;
; 6) No verification is made to insure the address parameters are valid (i.e. access
;    privileges). However, if adding the output buffer address to its' size causes
;    segment wrap, an invalid parameter error is returned.
;
;*********************************************************
;
; Stack Frame Equates
;
;*********************************************************

_OutBufSize   TEXTEQU <[EBP+28]>     ;output buffer size in bytes
_OutBufBase   TEXTEQU <[EBP+24]>     ;output buffer base address
_InpValBase   TEXTEQU <[EBP+20]>     ;input value base address

;                      [EBP+16]       PROC return address
;                      [EBP+12]       saved ESI contents
;                      [EBP+8]        saved EDI contents
;                      [EBP+4]        saved EBX contents
;                      [EBP]          saved EBP contents

_CheckSum     TEXTEQU <[EBP-4]>      ;checksum
_OutLastByte  TEXTEQU <[EBP-8]>      ;address of last output byte
_InpMaxAddr   TEXTEQU <[EBP-12]>     ;maximum input address

;*********************************************************
;
; Routine Initialization
;
;*********************************************************

;establish the stack frame

        push    esi                  ;save the registers
        push    edi
        push    ebx
        push    ebp
        mov     ebp,esp              ;EBP = stack frame base pointer

;verify valid output buffer parameters

        mov     edx,_OutBufSize      ;EDX = output buffer size
        xor     ecx,ecx              ;ECX = 0
        cmp     edx,7                ;minimum buffer size = 7 bytes
        jb      Error0               ;error - invalid parameter

        mov     eax,_OutBufBase      ;EAX = output buffer address
        mov     edi,eax              ;EDI = output buffer address
        dec     eax                  ;count to last byte
        add     eax,edx              ;EAX = address of last buffer byte
        jc      Error0               ;error - invalid parameter

;*********************************************************
;
; Initialize Registers and Variables for Base Conversion
;
;*********************************************************

;EAX = address of last output buffer byte
;ECX = 0
;EDI = output buffer base address

        push    ecx                  ;[EBP-4] = _CheckSum
        mov     esi,_InpValBase      ;ESI = input value base address
        push    eax                  ;[EBP-8] = _OutLastByte
        mov     edx,edi              ;EDX = active index
        mov     eax,esi              ;EAX = input value base address
        mov     [edi],cl             ;initialize the accumulator
        add     eax,1023             ;EAX = maximum input address
        mov     bx,3F0Ah             ;BL = 10, BH = 3Fh
        push    eax                  ;[EBP-12] = _InpMaxAddr
        jmp short Base03             ;ling long kai fang entry

;---------------------------------------------------------

;error - invalid parameter

Error0: push    87                   ;ERROR_INVALID_PARAMETER
        jmp short Error3             ;error exit

;---------------------------------------------------------

;*********************************************************
;
; Convert Base 10 ASCII to Base 64 Binary
; Ling Long Kai Fang Method
;
;*********************************************************

;EAX = accumulator register
; BL = 10 (multiplier)
; BH = 3Fh (AND mask)
; CL = carry register
; CH = 0
;EDX = active index
;ESI = source index
;EDI = destination index
;EBP = stack frame base pointer

Base00: mov     al,[edi]             ;load next accumulator byte

Base01: mul     bl                   ;multiply by 10
        add     eax,ecx              ;add the previous carry
        mov     ecx,eax              ;copy in ECX
        and     al,bh                ;AL = lower 6 bits
        mov     [edi],al             ;update the accumulator
        shr     ecx,6                ;CL = upper 4 bits for carry
        inc     edi                  ;point to next accumulator byte
        cmp     edi,edx              ;is this byte active ?
        jbe     Base00               ;yes - continue inner loop

        jecxz   Base02               ;no - any carry ?

        cmp     edi,_OutLastByte     ;yes - is it inside the buffer ?
        jae     Error2               ;error - output buffer overflow

        mov     [edi],cl             ;yes - update the accumulator
        mov     edx,edi              ;make this byte active
        mov     cl,ch                ;clear the old carry

Base02: mov     edi,_OutBufBase      ;point to 1st accumulator byte
        inc     esi                  ;point to next source byte
        cmp     esi,_InpMaxAddr      ;input too long ?
        jae     Error1               ;error - input too long

Base03: movzx   eax,byte ptr [esi]   ;load source byte
        xor     al,30h               ;make it decimal
        cmp     al,9                 ;is it valid ?
        ja      Base04               ;no - test for null

        add     _CheckSum,eax        ;add the decimal byte to checksum
        add     al,[edi]             ;add the accumulator
        cmp     ah,[esi+1]           ;is this the last digit ?
        jnz     Base01               ;no - continue outer loop

        jmp short Base06             ;yes - do the last digit

Base04: xor     al,30h               ;is it null ?
        jz      LCSum0               ;string length = 0 - do checksum

;---------------------------------------------------------

;error - invalid data

        push    13                   ;ERROR_INVALID_DATA
        jmp short Error3             ;error exit

;error - input too long

Error1: push    234                  ;ERROR_MORE_DATA (too long)
        jmp short Error3             ;error exit

;error - output buffer overflow

Error2: push    122                  ;ERROR_INSUFFICIENT_BUFFER

;error exit

;error code on stack

Error3: mov     edx,_OutBufBase      ;output buffer base address
        xor     ecx,ecx              ;zero length - null string
        pop     eax                  ;recall error code
        mov     [edx],cl             ;null terminator byte
        jmp     Rexit0               ;exit routine

;---------------------------------------------------------

;base conversion (cont.)

Base05: add     al,[edi]             ;add the accumulator

Base06: mov     ecx,eax              ;copy in ECX
        and     al,bh                ;AL = lower 6 bits
        shr     ecx,6                ;CL = upper 4 bits
        jz      Base07               ;no carry - last byte

        mov     [edi],al             ;update the accumulator
        inc     edi                  ;point to next accumulator byte
        xchg    eax,ecx              ;carry in AL for next pass
        cmp     edi,edx              ;is this byte active ?
        jbe     Base05               ;yes - continue loop

        mov     edx,edi              ;no - make it active

Base07: mov     [edi],al             ;update last accumulator byte
        mov     eax,esi              ;source pointer in EAX for length
        inc     eax                  ;adjust to end of string
        sub     eax,_InpValBase      ;EAX = input string length

;*********************************************************
;
; Append Length and Checksum Values to Base 64 String
;
;*********************************************************

;EAX = input string length
;EDX = active index

LCSum0: mov     ecx,edx              ;ECX = last base 64 byte address
        add     ecx,6                ;count checksums and terminator
        cmp     ecx,_OutLastByte     ;is there room ?
        ja      Error2               ;error - output buffer overflow

        add     _CheckSum,eax        ;_CheckSum = binary total + length
        mov     ebx,eax              ;copy in EBX
        inc     edx                  ;point to 1st checksum byte
        and     al,3Fh               ;lower 6 bits of length
        mov     [edx],al             ;L00,L01,L02,L03,L04,L05,0,0
        shr     ebx,4                ;position L06 to L09
        and     bl,3Ch               ;L06 to L09 only
        mov     eax,_CheckSum        ;EAX = checkusm bits
        inc     edx                  ;point to 2nd checksum byte
        mov     ecx,eax              ;ECX = checksum bits
        shr     eax,12               ;position C12 and C13
        or      al,bl                ;OR Cxx bits with Lxx bits
        mov     [edx],al             ;C12,C13,L06,L07,L08,L09,0,0
        mov     eax,ecx              ;EAX = checksum bits
        shr     eax,6                ;position C06 to C11
        inc     edx                  ;point to 3rd checksum byte
        and     eax,3Fh              ;C06 to C11 only
        mov     [edx],al             ;C06,C07,C08,C09,C10,C11,0,0
        inc     edx                  ;point to 4th checksum byte
        and     cl,3Fh               ;C00 to C05 only
        mov     [edx],cl             ;C00,C01,C02,C03,C04,C05,0,0

;*********************************************************
;
; Convert Base 64 Output String to Hashed ASCII
;
;*********************************************************

;EDX = address of last string byte (checksum low bits)

        mov     esi,offset _EncTbl   ;ESI - hash table address
        mov     edi,_OutBufBase      ;EDI = string base address
        xor     ebx,ebx              ;EBX = 0
        push    edi                  ;save string address for later
        mov     ecx,ebx              ;ECX = 0

Hash00: mov     al,[edi]             ;load source byte
        xor     cl,al                ;running total XOR checksum
        add     eax,ebx              ;incremental offset
        and     eax,3Fh              ;lower 6 bits only
        mov     al,[esi+eax]         ;get the hash byte
        inc     ebx                  ;next incremental offset
        mov     [edi],al             ;store destination byte
        inc     edi                  ;next string byte
        cmp     edi,edx              ;end of string ?
        jbe     Hash00               ;no - continue

        add     ecx,ebx              ;incremental offset
        and     ecx,3Fh              ;X00,X01,X02,X03,X04,X05,0,0
        mov     cl,[esi+ecx]         ;hashcode the XOR checksum
        pop     edx                  ;recall output string address
        mov     [edi],cl             ;append it to the string

;*********************************************************
;
; Set Return Values for Successful Conversion
;
;*********************************************************

;EAX < 256
;EDX = output string base address
;EDI = address of last output character

        inc     edi                  ;point to null terminator
        mov     al,ah                ;EAX = 0 = ERROR_SUCCESS
        mov     ecx,edi              ;ECX = address of null terminator
        mov     [edi],al             ;place a null terminator
        sub     ecx,edx              ;ECX = output string length

;*********************************************************
;
; Routine Exit
;
;*********************************************************

;EAX = return status
;ECX = output string length, not including null terminator
;EDX = output string address

Rexit0: leave                        ;same as MOV ESP,EBP - POP EBP
        pop     ebx                  ;restore registers
        pop     edi
        pop     esi
        ret     12                   ;discard 3 dwords and exit

Encod   ENDP

        OPTION  PROLOGUE:PrologueDef
        OPTION  EPILOGUE:EpilogueDef

;############################################################################################


;############################################################################################

        OPTION  PROLOGUE:NONE
        OPTION  EPILOGUE:NONE

Decod   PROC    szInputString:DWORD,lpOutputBuffer:DWORD,dwSize:DWORD
;
; Decode hashed base 64 ASCII string with checksums to decimal ASCII string
;
; accepts hashed base 64 ASCII string
; unhash/decode/expand to ASCII numeric string from 0 to 1023 bytes in length
; SUM checksum, XOR checksum, and length values are decoded and verified
; to be used with the companion Encod routine
;
; Version 1.02 - March, 2010
; by: Dave Sheldon - DedNDave
;
;*********************************************************
;
; Use:
;
;       INVOKE  Decod,
;               szInputString,
;               lpOutputBuffer,
;               dwSize
;
;   szInputString = address of zero terminated ASCII encoded string
;  lpOutputBuffer = address of output buffer
;          dwSize = size of output buffer in bytes
;
; Returns:
;
;  EAX = status code
;          0 = ERROR_SUCCESS
;              the string has successfully been decoded; ECX, EDX contain meaningful values
;         13 = ERROR_INVALID_DATA
;              indicates that a character was encountered that is not in the hash table or
;              the decoded XOR checksum, SUM checksum, and/or length are invalid
;              this may indicate that the string has been tampered with
;              it is possible, however, to tamper with a string and avoid detection
;         87 = ERROR_INVALID_PARAMETER
;              indicates a problem with the buffer address and/or size parameters
;        122 = ERROR_INSUFFICIENT_BUFFER
;              the output buffer is too small to accommodate the resultant decoded string
;        234 = ERROR_MORE_DATA
;              the input string is longer than the routine was designed to handle (572 bytes)
;
;  ECX = output string length, not including null terminator
;        if EAX = 0 (no error)
;           ECX = length of encoded string, not including null terminator
;        if EAX <> 0 (error)
;           ECX = 0
;
;  EDX = address of output buffer
;        if EAX = 0 (no error)
;           EDX = address of output buffer
;        if EAX <> 0 (error)
;           EDX = address of output buffer (null string)
;
; Notes:
;
; 1) The _EncTbl is required.
; 2) The original encoded string is not altered during operation.
; 3) The output buffer must include room for a null terminator.
; 4) The resulting output string is left-justified in the output buffer field.
; 5) The required size of the output buffer may be calculated as follows:
;
;    ********************************************************************
;    IMPORTANT: This formula is preliminary. It needs to be tested.
;               For now, use it as an ESTIMATE of the ouptut buffer size.
;    ********************************************************************
;
;    Dmax = Round((N-5) x Log10(64)) + 1
;    Log10(64) = 1.806179973983887171282433368347
;
;    where D = number of bytes in output buffer (including null terminator)
;          N = number of input bytes (not including null terminator)
;
;    Example: For an encoded input string of 50 characters...
;
;    Dmax = Round(45 x 1.806179973983887171282433368347) + 1
;         = 82
;    a 82 byte output buffer is required, which includes the null string terminator
;
; 6) No verification is made to insure the address parameters are valid (i.e. access
;    privileges). However, if adding the output buffer address to its' size causes
;    segment wrap, an invalid parameter error is returned.
;
;*********************************************************
;
; Stack Frame Equates
;
;*********************************************************

_OutBufSize   TEXTEQU <[EBP+28]>     ;output buffer size in bytes
_OutBufBase   TEXTEQU <[EBP+24]>     ;output buffer base address
_InpValBase   TEXTEQU <[EBP+20]>     ;input value base address

;                      [EBP+16]       PROC return address
;                      [EBP+12]       saved ESI contents
;                      [EBP+8]        saved EDI contents
;                      [EBP+4]        saved EBX contents
;                      [EBP]          saved EBP contents

_OutLastByte  TEXTEQU <[EBP-4]>      ;address of last output byte
_InpMaxAddr   TEXTEQU <[EBP-8]>      ;maximum input address
_DecodedLen   TEXTEQU <[EBP-12]>     ;decoded length
_DecodedSum   TEXTEQU <[EBP-16]>     ;decoded checksum

;*********************************************************
;
; Routine Initialization
;
;*********************************************************

;establish the stack frame

        push    esi                  ;save the registers
        push    edi
        push    ebx
        push    ebp
        mov     ebp,esp              ;EBP = stack frame base pointer

;verify valid output buffer parameters

        mov     edx,_OutBufSize      ;EDX = output buffer size
        cmp     edx,6                ;minimum buffer size = 6 bytes
        jb      Error0               ;error - invalid parameter

        mov     eax,_OutBufBase      ;EAX = output buffer address
        mov     esi,eax              ;ESI = output buffer address
        dec     eax                  ;count to last byte
        add     eax,edx              ;EAX = address of last buffer byte
        jc      Error0               ;error - invalid parameter

        push    eax                  ;[EBP-4] = _OutLastByte
        mov     edx,_InpValBase      ;EDX = input value base address
        mov     eax,edx              ;EAX = input value base address
        add     eax,572              ;EAX = maximum input address
        push    eax                  ;[EBP-8] = _InpMaxAddr

;*********************************************************
;
; Unhash the Input String and Verify XOR Checksum
;
;*********************************************************

;fill the table with 256 0FFh bytes

        mov     edi,offset _EncTbl   ;EDI = hash table address
        mov     ecx,64               ;ECX = 64 dword count
        or      eax,-1               ;EAX = 0FFFFFFFFh
        mov     ebx,ecx              ;EBX = ECX = count

Table0: push    eax                  ;push 64 dwords (256 bytes)
        dec     ebx                  ;count it
        jnz     Table0               ;loop until done

;set the 64 valid character locations

        inc     eax                  ;EAX = 0

Table1: mov     al,[edi]             ;load the hash table character
        inc     edi                  ;point to next hash character
        mov     [esp+eax],bl         ;set the table byte value
        inc     ebx                  ;next table byte value
        dec     ecx                  ;count it
        jnz     Table1               ;loop until done

;unhash the input string into the output buffer

        mov     ebx,ecx              ;EBX = ECX = 0
        jmp short UnHsh1             ;unhash loop entry

;---------------------------------------------------------

;error - invalid parameter

Error0: push    87                   ;ERROR_INVALID_PARAMETER
        jmp short Error2             ;error exit

;---------------------------------------------------------

UnHsh0: mov     al,[esp+eax]         ;unhash/translate
        cmp     al,0FFh              ;is it a valid character ?
        jz      Error1               ;error - invalid data

        sub     al,cl                ;incremental offset
        and     al,3Fh               ;lower 6 bits only
        inc     ecx                  ;next incremental offset
        xor     ebx,eax              ;XOR checksum
        inc     edx                  ;point to next source byte
        mov     [esi],al             ;store the unhashed byte
        inc     esi                  ;point to the next destination
        cmp     edx,_InpMaxAddr      ;max length exceeded ?
        ja      Error3               ;error - input too long

        cmp     esi,_OutLastByte     ;is it inside the buffer ?
        ja      Error4               ;error - output buffer overflow

UnHsh1: mov     al,[edx]             ;load source string byte
        or      al,al                ;null terminator ?
        jnz     UnHsh0               ;no - unhash it

        add     esp,256              ;yes - discard the hash table

;verify unhashed XOR checksum

        or      ebx,ebx              ;XOR checksum valid ?
        jnz     Error1               ;error - invalid data

;*********************************************************
;
; Extract the Decoded Checksum and Length Values
;
;*********************************************************

        dec     esi                  ;skip the XOR checksum byte
        dec     esi                  ;point to 4th checksum byte
        movzx   ecx,byte ptr [esi]   ;C00,C01,C02,C03,C04,C05,0,0
        dec     esi                  ;point to 3rd checksum byte
        mov     al,[esi]             ;C06,C07,C08,C09,C10,C11,0,0
        shl     eax,6                ;position C06 to C11
        dec     esi                  ;point to 2nd checksum byte
        or      ecx,eax              ;C00 to C11 bits
        movzx   eax,byte ptr [esi]   ;C12,C13,L06,L07,L08,L09,0,0
        mov     edx,eax              ;copy in EDX
        and     al,3                 ;C12 and C13 bits
        dec     esi                  ;point to 1st checksum byte
        shl     eax,12               ;position C12 and C13
        and     dl,3Ch               ;L06 to L09 bits
        or      ecx,eax              ;ECX = checksum + length
        shl     edx,4                ;position L06 to L09
        or      dl,[esi]             ;L00,L01,L02,L03,L04,L05,0,0
        cmp     edx,_OutBufSize      ;is length < buffer size ?
        jae     Error4               ;error - output buffer overflow

        sub     ecx,edx              ;subtract the decoded length
        push    edx                  ;[EBP-12] = _DecodedLen
        dec     esi                  ;point to 6th checksum byte
        push    ecx                  ;[EBP-16] = _DecodedSum

;*********************************************************
;
; Initialize Registers for Base Conversion
;
;*********************************************************

;ESI = address of high order base 64 input byte
;EBP = stack frame base pointer

        mov     ebx,_OutLastByte     ;EBX = active index
        xor     ecx,ecx              ;ECX = 0
        mov     [ebx],cl             ;initialize the accumulator
        jmp short Base03             ;ling long kai fang entry

;---------------------------------------------------------

;error - invalid data

Error1: push    13                   ;ERROR_INVALID_DATA

Error2: jmp short Error6             ;error exit

;error - input too long

Error3: push    234                  ;ERROR_MORE_DATA (too long)
        jmp short Error6             ;error exit

;error - output buffer overflow

Error4: push    122                  ;ERROR_INSUFFICIENT_BUFFER
        jmp short Error6             ;error exit

;---------------------------------------------------------

;*********************************************************
;
; Convert Base 64 Binary to Base 100
; Ling Long Kai Fang Method
;
;*********************************************************

;EBX = active index
;ECX = carry
;ESI = input value index
;EDI = accumulator index

Base00: movzx   eax,byte ptr [edi]   ;load next accumulator byte

Base01: shl     eax,6                ;multiply by 64
        add     eax,ecx              ;add the previous carry
        mov     edx,28F5C29h         ;multiplier = 2^32 / 100
        mul     edx                  ;multiply to divide
        mov     ecx,edx              ;ECX = carry
        mov     edx,100              ;multiplier = 100
        mul     edx                  ;extend the remainder
        mov     [edi],dl             ;update the accumulator
        dec     edi                  ;point to next accumulator byte
        cmp     edi,ebx              ;is this byte active ?
        jae     Base00               ;yes - continue inner loop

        jecxz   Base02               ;no - any carry ?

        cmp     edi,esi              ;yes - is it inside the buffer ?
        jb      Error4               ;error - output buffer overflow

        mov     [edi],cl             ;yes - update the accumulator
        mov     ebx,edi              ;make this byte active
        mov     cl,ch                ;clear the old carry

Base02: dec     esi                  ;point to next source byte

Base03: movzx   eax,byte ptr [esi]   ;load source byte
        mov     edi,_OutLastByte     ;point to 1st accumulator byte
        add     al,[edi]             ;add the accumulator
        cmp     esi,_OutBufBase      ;is this the last digit ?
        ja      Base01               ;no - continue outer loop

        push    100                  ;100 on the stack
        inc     ecx                  ;ECX = 1
        pop     edx                  ;EDX = 100
        jmp short Base05             ;do the last digit

Base04: add     al,[edi]             ;add the accumulator

Base05: cmp     eax,edx              ;any carry ?
        jb      Base06               ;no - last byte

        sub     eax,edx              ;yes - subtract 100
        mov     [edi],al             ;update the accumulator
        dec     edi                  ;point to next accumulator byte
        mov     eax,ecx              ;carry in AL for next pass
        cmp     edi,ebx              ;is this byte active ?
        jae     Base04               ;yes - continue loop

        mov     ebx,edi              ;no - make it active

Base06: mov     [edi],al             ;update last accumulator byte

;*********************************************************
;
; Calculate String Length
;
;*********************************************************

;EAX < 256
;EBX = active index
;ECX = 1
;ESI = _OutBufBase

        mov     al,[ebx]             ;EAX = high order byte
        mov     edi,_OutLastByte     ;EDI = low order byte address
        aam                          ;split into AH/AL
        sub     edi,ebx              ;EDI = byte count less 1
        cmp     eax,ecx              ;set CF if AL = AH = 0
        mov     edx,_DecodedLen      ;EDX = decoded string length
        sbb     cl,ch                ;if AX = 0, CL = 0 else CL = 1
        shl     edi,1                ;EDI = digit count less AX
        cmp     ch,ah                ;set CF of AH > 0
        adc     cl,ch                ;ECX = signifigant digits in AX
        add     edi,ecx              ;EDI = total signifigant digits
        sub     edx,edi              ;EDX = leading zero count
        jnc     LZero0               ;valid if not negative

;---------------------------------------------------------

;error - invalid data

Error5: push    13                   ;ERROR_INVALID_DATA

;error exit

;error code on stack

Error6: mov     edx,_OutBufBase      ;output buffer base address
        xor     ecx,ecx              ;zero length - null string
        pop     eax                  ;recall error code
        mov     [edx],cl             ;null terminator byte
        jmp short Rexit0             ;exit routine

;---------------------------------------------------------

;*********************************************************
;
; Place Any Required Leading Zeros
;
;*********************************************************

;EAX = first 0, 1 or 2 binary digits in AH, AL
;EBX = active index
;ECX = AX digit count (0, 1 or 2)
;EDX = leading zero count
;ESI = _OutBufBase
;EDI = signifigant digit count including in-register digits

LZero0: or      edx,edx              ;any leading zeros ?
        push    esi                  ;save output string address
        jz      Digit0               ;no - skip leading zeros

        push    eax                  ;yes - save in-register digits
        mov     al,30h               ;ASCII zero

LZero1: mov     [esi],al             ;place a leading zero
        inc     esi                  ;point to next buffer byte
        dec     edx                  ;count it
        jnz     LZero1               ;loop until done

        pop     eax                  ;recall in-register digits

;*********************************************************
;
; Place the In-Register Digits
;
;*********************************************************

;EAX = first 0, 1 or 2 binary digits in AH, AL
;EBX = active index
;ECX = AX digit count (0, 1 or 2)
;EDX = 0
;ESI = output buffer index
;EDI = signifigant digit count including in-register digits

Digit0: jecxz   Ascii0               ;skip in-register digits

        sub     edi,ecx              ;adjust total count
        dec     ecx                  ;only 1 digit ?
        jz      Digit1               ;yes - do 1 digit

        mov     dl,ah                ;place in DL for checksum
        or      ah,30h               ;make it ASCII
        mov     [esi],ah             ;store the digit
        sub     _DecodedSum,edx      ;subtract from checksum
        inc     esi                  ;point to next buffer byte

Digit1: mov     dl,al                ;place in DL for checksum
        or      al,30h               ;make it ASCII
        mov     [esi],al             ;store the digit
        sub     _DecodedSum,edx      ;subtract from checksum
        inc     esi                  ;point to next buffer byte

;*********************************************************
;
; Convert Remaining Digits to ASCII
;
;*********************************************************

;EBX = active index
;EDX < 256
;EDI = signifigant digit count

Ascii0: shr     edi,1                ;byte count = digit count / 2
        jz      Term00               ;no more digits - terminate

        inc     ebx                  ;already did this byte

Ascii1: mov     al,[ebx]             ;load next 2-digit byte
        aam                          ;split into AH/AL
        inc     ebx                  ;point to next byte
        mov     dl,al                ;place in DL for checksum
        xchg    al,ah                ;swap digit locations
        add     dl,al                ;both digits for checksum
        or      ax,3030h             ;make them ASCII
        sub     _DecodedSum,edx      ;subtract from checksum
        mov     [esi],ax             ;store 2 digits
        inc     esi                  ;point to next buffer byte
        inc     esi                  ;point to next buffer byte
        dec     edi                  ;count it
        jnz     Ascii1               ;loop until done

;*********************************************************
;
; Set Return Values for Successful Conversion
;
;*********************************************************

;EDI = 0

Term00: cmp     edi,_DecodedSum      ;_DecodedSum correct ?
        jnz     Error5               ;error - invalid data

        mov     ecx,_DecodedLen      ;ECX = output string length
        xchg    eax,edi              ;EAX = 0 = ERROR_SUCCESS
        pop     edx                  ;recall output string address
        mov     [edx+ecx],al         ;place a null terminator

;*********************************************************
;
; Routine Exit
;
;*********************************************************

;EAX = return status
;ECX = output string length, not including null terminator
;EDX = output string address

Rexit0: leave                        ;same as MOV ESP,EBP - POP EBP
        pop     ebx                  ;restore registers
        pop     edi
        pop     esi
        ret     12                   ;discard 3 dwords and exit

Decod   ENDP

        OPTION  PROLOGUE:PrologueDef
        OPTION  EPILOGUE:EpilogueDef

;############################################################################################

;end of include file

